@@ -30,7 +30,7 @@ fn real_main() -> i32 {
             }
         }
 
-        if (&*file.name()).ends_with('/') {
+        if (*file.name()).ends_with('/') {
             println!("File {} extracted to \"{}\"", i, outpath.display());
             fs::create_dir_all(&outpath).unwrap();
         } else {
@@ -59,5 +59,6 @@ fn real_main() -> i32 {
             }
         }
     }
-    return 0;
+
+    0
 }
